/*Copyright © 2013 Alexander Kobzar (sashakobzar@gmail.com)

GETALP
Grenoble Informatics Laboratory
France

This script allows you to generate a Blast annotation file based on a source text and its Moses-generated translation with the help of the word-to-word alignment information*/
using System;
using System.IO;
using System.Linq;
using System.Collections;
using System.Collections.Generic;
using System.Text.RegularExpressions;

class mwe2blast
{
	//Replace HTML and XML entities by characters
	public static string replaceEntity (string str)
	{
		return System.Web.HttpUtility.HtmlDecode (str.Replace (@"&apos;", "'").Replace (@"&quot;", @"""").Replace (@"&amp;", "&").Replace (@"&lt;", "<").Replace (@"&gt;", ">"));
	}

	public static void closeReaders (params StreamReader[] readers)
	{
		foreach (StreamReader stReader in readers)
			stReader.Close ();
	}

	public static void Main (string[] args)
	{
		string sourceFile, translFile, alnFile, mweBlastFile, categFile, blastFile, blastAlmSource, positions, blastSent, strCateg = "", blastCateg = "", strLine, alnString;
		int sentCount = 0;
		byte shift = 0;
		short parseAln;
		bool defcatExist = false;
		string[] wordPos;
		StreamReader readMWEBlast, readTransl, readAlm, readSource, readCategFile;
		StreamWriter writeBlast;
		Dictionary<string, List<string>> dictAlm;
		Dictionary<string, string> sentMWEPos = new Dictionary<string, string> ();
		string message = @"mwe2blast.exe <source text file> <translated text file> <Moses-generated alignment file> <mwe2blast file> <Blast category file> <output file> [OPTIONS]

Options:
--------
-defanot <level-item>. Specifies the annotation which will be applied to all the sentences in the output. Otherwise, sentences will be marked with the first appearance of 'undef' in the specified category file or with the last element of the last level.

-shift <number>. -shift <number>. Specifies a shift applied to MWE positions in the mwe2blast file in order to match the ones in the alignment information file generated by Moses. For hierarchical systems, <-shift 1> may be needed.";
		#region Process command-line arguments
		if (args.Count () != 0 && (args [0] == "/?" || args [0] == "-help")) {
			Console.WriteLine (message);
			return;
		}
		if (args.Count () < 6) {
			Console.WriteLine ("You did not specify all required parameters.\r\nType '[mono] filterblast.exe -help' or '[mono] filterblast.exe /?' for help.");
			return;
		}
		//Check whether all specified files exist
		for (int i = 0; i <= 4; i++) {
			if (!File.Exists (args [i])) {
				Console.WriteLine ("{0} does not exists.", args [i]);
				return;
			}
		}
		sourceFile = args [0];
		translFile = args [1];
		alnFile = args [2];
		mweBlastFile = args [3];
		categFile = Path.GetFileName (args [4]);
		blastFile = args [5] + ".blast";
		readMWEBlast = new StreamReader (mweBlastFile);
		readTransl = new StreamReader (translFile);
		readAlm = new StreamReader (alnFile);
		readSource = new StreamReader (sourceFile);
		try {
			writeBlast = new StreamWriter (blastFile, false);
		} catch (Exception e) {
			Console.WriteLine ("The Blast file you specified, {0}, cannot be created due to the following reason:\r\n{1}", blastFile, e.Message);
			return;
		}
		readCategFile = new StreamReader (categFile);
		//Process command-line options
		for (int i = 6; i < args.Count(); i++) {
			switch (args [i]) {
			case "-shift":
				if (i < args.Count () - 1) {
					if (!byte.TryParse (args [Array.IndexOf (args, "-shift") + 1], out shift)) {
						Console.WriteLine ("-shift is specified incorrectly. Please, provide an integer.");
						return;
					}
				} else {
					Console.WriteLine ("Shift is missing. Please, provide an integer.");
					return;
				}
				i++;
				break;
			case "-defanot":
				if (i < args.Count () - 1) {
					blastCateg = args [Array.IndexOf (args, "-defanot") + 1];
					if (!Regex.IsMatch (blastCateg, "\\A[^#,\\- ]+-[^#,\\- ]+\\Z")) {
						Console.WriteLine ("The default category should be specified in the following format: level-item");
						return;
					}
				} else {
					Console.WriteLine ("Default category is missing.\r\nType '[mono] filterblast.exe -help' or '[mono] filterblast.exe /?' for help.");
					return;
				}
				i++;
				break;
			default:
				Console.WriteLine ((args [i].StartsWith ("-") ? "Invalid option: " : "Invalid argument: ") + args [i] + "\r\nType '[mono] filterblast.exe -help' or '[mono] filterblast.exe /?' for help.");
				return;
			}
		}
		#endregion
		#region Determine default annotation
		//Search for the default category		
		while (((strLine = readCategFile.ReadLine()) != null)) {
			strLine = strLine.Trim ();
			if (strLine == "") {
				continue;
			}
			if (readCategFile.ReadLine () == "list") {
				strCateg = readCategFile.ReadLine ().ToLower ().Trim ();
				//Search for the first occurrence of 'undef'
				if ((!args.Contains ("-defanot")) && (Regex.IsMatch (strCateg, ",undef(,| |\\Z)"))) {
					strCateg = strLine + "-undef";
					break;
				} else {
					//Check whether the specified default category exists
					if ((args.Contains ("-defanot")) && (strLine == blastCateg.Substring (0, blastCateg.IndexOf ('-'))) && (Regex.IsMatch (strCateg, "," + blastCateg.Substring (blastCateg.IndexOf ('-') + 1) + "(,| |\\Z)"))) {
						defcatExist = true;
						break;
					}
					//Find default category
					strCateg = strLine + "-" + Regex.Match (strCateg, "(?<=,)[^, ]+(?=(,[^, ]*)?\\s*\\Z)").Value;
				}
			} else {
				readCategFile.ReadLine ();
			}
		}
		readCategFile.Close ();
		//Check whether the default category is valid
		if (args.Contains ("-defanot")) {
			if (!defcatExist) {
				Console.WriteLine ("There is no {0} category in {1}.", blastCateg, categFile);
				return;
			}
		} else {
			if (!Regex.IsMatch (strCateg, "\\A[^#,\\- ]+-[^#,\\- ]+\\Z")) {
				Console.WriteLine ("The category file you specified, {0}, is malformed.", categFile);
				return;
			} else {
				blastCateg = strCateg;
			}
		}
		//No 'undef' category found
		if (!args.Contains ("-defanot")) {
			if (!blastCateg.EndsWith ("-undef")) {
				Console.WriteLine ("There is no 'undef' category in " + categFile + ". All sentences will be marked " + blastCateg + "\r\nDo you want to continue (y/n)?");
				while (true) {
					strLine = Console.ReadLine ().Trim ().ToLower ();
					if (strLine == "y") {
						strCateg = blastCateg;
						break;
					}
					if (strLine == "n") {
						return;
					}
					Console.WriteLine ("Please type 'y' or 'n'");
				}
			} else {
				Console.WriteLine ("All sentences will carry the following annotation: {0}", blastCateg);
			}
		}
		#endregion
		//Extract MWE positions in each sentence	
		while ((positions = readMWEBlast.ReadLine()) != null) {
			strLine = readMWEBlast.ReadLine ();
			if (strLine == null || positions.Trim () == "") {
				Console.WriteLine ("The specified MWEBlast file, {0}, contains an error. Perhaps, there is an empty string in the file.", mweBlastFile);
				closeReaders (readAlm, readSource, readTransl, readMWEBlast);
				writeBlast.Close ();
				return;
			}
			if (!sentMWEPos.ContainsKey (strLine))
				sentMWEPos.Add (strLine, positions);
		}
		writeBlast.WriteLine ("#Sentencetypes src sys\r\n#catfile " + categFile);
		//Create a blast file
		while (readSource.Peek() != -1) {
			//Extract word-to-word alignments
			dictAlm = new Dictionary<string, List<string>> ();
			wordPos = null;
			#region Error cheking
			if (readAlm.Peek () == -1) {
				Console.WriteLine ("{0} has less strings (sentences) than {1}", alnFile, sourceFile);
				closeReaders (readAlm, readSource, readTransl, readMWEBlast);
				writeBlast.Close ();
				return;
			}
			#endregion
			alnString = readAlm.ReadLine ();
			foreach (string strAln in alnString.Trim().Split(new char[] { ' ' })) {
				wordPos = strAln.Split (new char[] { '-' });
				#region Error checking in alignment file
				if (alnString.Trim () == "") {
					Console.WriteLine ("WARNING! {0} contains an empty string.", alnFile);
					continue;
				} else if (wordPos.GetLength (0) < 2 || wordPos.GetLength (0) > 2 || (!short.TryParse (wordPos [0], out parseAln) || !short.TryParse (wordPos [1], out parseAln))) {
					Console.WriteLine ("The following line in the alignment file you specified, {0}, is malformed:\r\n{1}", alnFile, alnString);
					closeReaders (readAlm, readSource, readTransl, readMWEBlast);
					writeBlast.Close ();
					return;
				}
				#endregion
				if (!dictAlm.ContainsKey (wordPos [0]))
					dictAlm.Add (wordPos [0], new List<string> () { wordPos[1] });
				else
					dictAlm [wordPos [0]].Add (wordPos [1]);
			}
			strLine = readSource.ReadLine ();
			#region Error checking
			if (strLine == "") {
				Console.WriteLine ("WARNING! {0} contains an empty string.", sourceFile);
				readAlm.ReadLine ();
				readTransl.ReadLine ();
				continue;
			}
			if (!sentMWEPos.ContainsKey (strLine)) {
				Console.WriteLine ("The following sentence is not contained in {1}:\r\n{0}\r\nPerhaps, there is an empty string in the file.", strLine, mweBlastFile);
				closeReaders (readAlm, readSource, readTransl, readMWEBlast);
				writeBlast.Close ();
				return;
			}
			positions = sentMWEPos [strLine].Trim ();
			if (readTransl.Peek () == -1) {
				Console.WriteLine ("{0} has less strings (sentences) than {1}", translFile, sourceFile);
				closeReaders (readAlm, readSource, readTransl, readMWEBlast);
				writeBlast.Close ();
				return;
			}
			blastSent = replaceEntity (strLine) + "\r\n\r\n" + replaceEntity ((strLine = readTransl.ReadLine ())) + "\r\n\r\n";
			if (strLine == "") {
				Console.WriteLine ("WARNING! {0} contains an empty string.", translFile);
			}
			#endregion
			//Extract alignments for each MWE in the current sentence
			foreach (string mwePos in positions.Split(new char[] { '\t' })) {
				blastAlmSource = "";
				//Source sentence MWE
				foreach (string pos in mwePos.Split(new char[] { ' ' })) {
					blastAlmSource += blastAlmSource == "" ? pos : "," + pos;
				}
				blastAlmSource += "#";
				//Extract positions of all words from the target sentence aligned with the MWE in the current source one 		
				foreach (string[] targetAlm in dictAlm.Where(x => mwePos.Split(new char[] { ' ' }).Contains((Convert.ToInt16(x.Key) + shift).ToString())).Select(x => x.Value.ToArray())) {
					//Extract positions of all words aligned with the current word 
					foreach (string pos in targetAlm) {
						blastAlmSource += blastAlmSource.EndsWith ("#") ? (Convert.ToInt16 (pos) + shift).ToString () : "," + (Convert.ToInt16 (pos) + shift).ToString ();
					}
				}
				blastAlmSource += "#-1#" + blastCateg;
				writeBlast.WriteLine (blastSent + blastAlmSource);
				sentCount++;
			}
		}
		#region Error checking
		writeBlast.Close ();
		if (readAlm.Peek () != -1) {
			Console.WriteLine ("{0} has more strings (sentences) than {1}", alnFile, sourceFile);
			closeReaders (readAlm, readSource, readTransl, readMWEBlast);
			return;
		}
		if (readTransl.Peek () != -1) {
			Console.WriteLine ("{0} has more strings (sentences) than {1}", translFile, sourceFile);
			closeReaders (readAlm, readSource, readTransl, readMWEBlast);
			return;
		}
		#endregion
		closeReaders (readAlm, readSource, readTransl, readMWEBlast);
		Console.WriteLine ("{0} contains {1} sentences to judge.", blastFile, sentCount);
	}
}